import dataclasses
import inspect
import json
from collections import OrderedDict
from typing import Any, Dict, List, Literal, Optional, Set, Type, TypeVar, Union, overload

from dacite import Config, from_dict
from fastavro.validation import validate

from . import case, serialization
from .dacite_config import generate_dacite_config
from .fields.base import Field
from .parser import Parser
from .types import JsonDict
from .utils import UserDefinedType, standardize_custom_type

_schemas_cache: Dict["Type[AvroModel]", dict] = {}
_dacite_config_cache: Dict["Type[AvroModel]", Config] = {}
TSelf = TypeVar("TSelf", bound="AvroModel")


class AvroModel:
    _parser: Optional[Parser] = None
    _parent: Optional[Type["AvroModel"]] = None
    _user_defined_types: Set[UserDefinedType] = set()
    _rendered_schema: OrderedDict = dataclasses.field(default_factory=OrderedDict)

    @classmethod
    def get_fullname(cls: Type["AvroModel"]) -> str:
        """
        Fullname is composed of two parts: a name and a namespace
        separated by a dot. A namespace is a dot-separated sequence of such names.
        The empty string may also be used as a namespace to indicate the null namespace.
        Equality of names (including field names and enum symbols)
        as well as fullnames is case-sensitive.
        """
        # we need to make sure that the schema has been generated
        cls.generate_schema()
        assert cls._parser
        metadata = cls._parser.metadata

        if metadata.namespace:
            # if the current record has a namespace we use it
            return f"{metadata.namespace}.{cls.__name__}"
        elif cls._parent is not None:
            # if the record has a parent then we try to use the parent namespace
            assert cls._parent._parser
            parent_metadata = cls._parent._parser.metadata
            if parent_metadata.namespace:
                return f"{parent_metadata.namespace}.{cls.__name__}"
        return cls.__name__

    @classmethod
    def generate_schema(
        cls: Type["AvroModel"], schema_type: serialization.SerializationType = "avro"
    ) -> Optional[OrderedDict]:
        if cls._parser is None:
            # let's live open the possibility to define different
            # schema definitions like json
            if schema_type == "avro":
                # cache the schema
                cls._parser = cls._generate_parser()
                cls._rendered_schema = cls._parser.render()
            else:
                raise ValueError("Invalid type. Expected avro schema type.")

        return cls._rendered_schema

    @classmethod
    def _get_serialization_context(cls) -> JsonDict:
        """
        Returns:
            Dict[str, Any] with the necesary context for the serialization/deserialization process
            It contains at least all the AvroModel defined by the end users represented by
            AvroModel.__name__: AvroModel entries
        """
        return {user_type.model.__name__: user_type.model for user_type in cls._user_defined_types}

    @classmethod
    def _generate_parser(cls: Type["AvroModel"]) -> Parser:
        return Parser(type=cls, parent=cls._parent or cls)

    @classmethod
    def avro_schema(cls: Type["AvroModel"], case_type: Optional[str] = None, **kwargs) -> str:
        return json.dumps(cls.avro_schema_to_python(case_type=case_type), **kwargs)

    @classmethod
    def avro_schema_to_python(
        cls: Type["AvroModel"],
        parent: Optional[Type["AvroModel"]] = None,
        case_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        if parent is not None:
            # in this case the current class is a child with a parent
            # we recalculate the schema definition to prevent re usages
            cls._parent = parent
            cls._parser = None
        else:
            # This happens when an AvroModel is the root of the tree (first class in the hierarchy)
            # Because intermediate schemas can be reused as a root later, we need to reset them
            # Example with A as a root:
            #     A -> B -> C -> D
            #
            # After generating the A.avro_schema the parent of B is A,
            # if we want to do B.avro_schema (now B is the root)
            # B should clean the data that was only valid when it was the child
            cls._reset_parser()

        avro_schema = cls.generate_schema()

        if case_type is not None:
            avro_schema = case.case_record(cls._rendered_schema, case_type)  # type: ignore

        return json.loads(json.dumps(avro_schema))

    @classmethod
    def get_fields(cls: Type["AvroModel"]) -> List[Field]:
        if cls._parser is None:
            cls.generate_schema()
        return cls._parser.fields  # type: ignore

    @classmethod
    def _reset_parser(cls: Type["AvroModel"]) -> None:
        """
        Reset all the values to original state.
        """
        cls._user_defined_types = set()
        cls._parser = None
        cls._parent = None

    @classmethod
    @overload
    def deserialize(
        cls: Type[TSelf],
        data: bytes,
        serialization_type: serialization.SerializationType = ...,
        create_instance: Literal[True] = ...,
        writer_schema: Optional[Union[JsonDict, Type["AvroModel"]]] = ...,
    ) -> TSelf: ...
    @classmethod
    @overload
    def deserialize(
        cls: Type[TSelf],
        data: bytes,
        serialization_type: serialization.SerializationType = ...,
        create_instance: Literal[False] = ...,
        writer_schema: Optional[Union[JsonDict, Type["AvroModel"]]] = ...,
    ) -> JsonDict: ...
    @classmethod
    @overload
    def deserialize(
        cls: Type[TSelf],
        data: bytes,
        serialization_type: serialization.SerializationType = ...,
        create_instance: bool = ...,
        writer_schema: Optional[Union[JsonDict, Type["AvroModel"]]] = ...,
    ) -> Union[TSelf, JsonDict]: ...
    @classmethod
    def deserialize(
        cls,
        data,
        serialization_type="avro",
        create_instance=True,
        writer_schema=None,
    ):
        payload = cls.deserialize_to_python(data, serialization_type, writer_schema)
        obj = cls.parse_obj(payload)

        if not create_instance:
            return obj.to_dict()
        return obj

    @classmethod
    def deserialize_to_python(  # This can be used straight with a pydantic dataclass to bypass dacite
        cls: Type["AvroModel"],
        data: bytes,
        serialization_type: serialization.SerializationType = "avro",
        writer_schema: Union[JsonDict, Type["AvroModel"], None] = None,
    ) -> dict:
        if inspect.isclass(writer_schema) and issubclass(writer_schema, AvroModel):
            # mypy does not understand redefinitions
            writer_schema: JsonDict = writer_schema.avro_schema_to_python()  # type: ignore

        schema = _schemas_cache.get(cls)
        if schema is None:
            schema = cls.avro_schema_to_python()
            _schemas_cache[cls] = schema
        return serialization.deserialize(
            data=data,
            schema=schema,
            serialization_type=serialization_type,
            context=cls._get_serialization_context(),
            writer_schema=writer_schema,  # type: ignore
        )

    @classmethod
    def parse_obj(cls: Type[TSelf], data: Dict) -> TSelf:
        config = _dacite_config_cache.get(cls)
        if config is None:
            config = generate_dacite_config(cls)
            _dacite_config_cache[cls] = config
        return from_dict(data_class=cls, data=data, config=config)

    @classmethod
    def fake(cls: Type["AvroModel"], **data: Any) -> "AvroModel":
        """
        Creates a fake instance of the model.

        Keyword Arguments:
            Any user values to use in the instance. All fields not explicitly passed
            will be filled with fake data.
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return from_dict(data_class=cls, data=payload, config=generate_dacite_config(cls))

    def asdict(self) -> JsonDict:
        return {
            field.name: standardize_custom_type(
                field_name=field.name, value=getattr(self, field.name), model=self, base_class=AvroModel
            )
            for field in dataclasses.fields(self)  # type: ignore[arg-type]
        }

    def serialize(self, serialization_type: serialization.SerializationType = "avro") -> bytes:
        klass = type(self)
        schema = _schemas_cache.get(klass)
        if schema is None:
            schema = self.avro_schema_to_python()
            _schemas_cache[klass] = schema

        return serialization.serialize(
            self.asdict(),
            schema,
            serialization_type=serialization_type,
        )

    def validate(self) -> bool:
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    def to_dict(self) -> JsonDict:
        return dataclasses.asdict(self)  # type: ignore

    def to_json(self, **kwargs: Any) -> str:
        data = serialization.to_json(self.to_dict())
        return json.dumps(data, **kwargs)



from .fields.field_utils import (
    ARRAY,
    BOOLEAN,
    BYTES,
    DATE,
    DECIMAL,
    DOUBLE,
    ENUM,
    FIXED,
    FLOAT,
    INT,
    LOGICAL_DATE,
    LOGICAL_DATETIME_MICROS,
    LOGICAL_DATETIME_MILIS,
    LOGICAL_TIME_MICROS,
    LOGICAL_TIME_MILIS,
    LOGICAL_UUID,
    LONG,
    MAP,
    NULL,
    PYTHON_TYPE_TO_AVRO,
    RECORD,
    STRING,
    TIME_MICROS,
    TIME_MILLIS,
    TIMESTAMP_MICROS,
    TIMESTAMP_MILLIS,
    UUID,
)
from .fields.fields import (
    AvroField,
    BooleanField,
    BytesField,
    ContainerField,
    DateField,
    DatetimeField,
    DatetimeMicroField,
    DecimalField,
    DictField,
    DoubleField,
    EnumField,
    FixedField,
    FloatField,
    ImmutableField,
    IntField,
    ListField,
    LiteralField,
    LongField,
    NoneField,
    RecordField,
    SelfReferenceField,
    StringField,
    TimeMicroField,
    TimeMilliField,
    TupleField,
    UnionField,
    UUIDField,
)
from .main import AvroModel
from .model_generator.generator import BaseClassEnum, ModelGenerator, ModelType
from .serialization import AVRO, AVRO_JSON, SerializationType, deserialize, serialize
from .types import DateTimeMicro, Float32, Int32, TimeMicro, condecimal, confixed

__all__ = [
    "AvroModel",
    "BaseClassEnum",
    "ModelType",
    "ModelGenerator",
    "Int32",
    "Float32",
    "TimeMicro",
    "DateTimeMicro",
    "condecimal",
    "confixed",
    "BOOLEAN",
    "NULL",
    "INT",
    "FLOAT",
    "LONG",
    "DOUBLE",
    "BYTES",
    "STRING",
    "ARRAY",
    "ENUM",
    "MAP",
    "FIXED",
    "DATE",
    "UUID",
    "DECIMAL",
    "RECORD",
    "TIME_MILLIS",
    "TIME_MICROS",
    "TIMESTAMP_MILLIS",
    "TIMESTAMP_MICROS",
    "LOGICAL_DATE",
    "LOGICAL_TIME_MILIS",
    "LOGICAL_TIME_MICROS",
    "LOGICAL_DATETIME_MILIS",
    "LOGICAL_DATETIME_MICROS",
    "LOGICAL_UUID",
    "PYTHON_TYPE_TO_AVRO",
    "ImmutableField",
    "StringField",
    "IntField",
    "LongField",
    "BooleanField",
    "DoubleField",
    "FloatField",
    "BytesField",
    "NoneField",
    "ContainerField",
    "ListField",
    "TupleField",
    "DictField",
    "UnionField",
    "LiteralField",
    "FixedField",
    "EnumField",
    "SelfReferenceField",
    "DateField",
    "DatetimeField",
    "DatetimeMicroField",
    "TimeMilliField",
    "TimeMicroField",
    "UUIDField",
    "DecimalField",
    "RecordField",
    "AvroField",
    "AVRO",
    "AVRO_JSON",
    "SerializationType",
    "serialize",
    "deserialize",
]



import dataclasses
import enum
import typing
from datetime import datetime, timezone
from functools import lru_cache

import typing_extensions
from typing_extensions import Annotated, get_origin

from .protocol import ModelProtocol  # pragma: no cover
from .types import FieldInfo, JsonDict, UnionType

if typing.TYPE_CHECKING:
    from dataclasses_avroschema import AvroModel  # pragma: no cover


try:
    import pydantic  # pragma: no cover
    from pydantic import v1
except ImportError:  # type: ignore # pragma: no cover
    pydantic = None  # type: ignore # pragma: no cover


try:
    import faust  # pragma: no cover
except ImportError:  # type: ignore # pragma: no cover
    faust = None  # type: ignore # pragma: no cover


@lru_cache(maxsize=None)
def _get_typing_objects_by_name_of(name: str) -> tuple[typing.Any, ...]:
    """Get the member named `name` from both `typing` and `typing-extensions` (if it exists)."""
    result = tuple(getattr(module, name) for module in (typing, typing_extensions) if hasattr(module, name))
    if not result:
        raise ValueError(f"Neither `typing` nor `typing_extensions` has an object called {name!r}")
    return result


def _is_typing_name(obj: object, name: str) -> bool:
    """Return whether `obj` is the member of the typing modules (includes the `typing-extensions` one) named `name`."""
    # Using `any()` is slower:
    for thing in _get_typing_objects_by_name_of(name):
        if obj is thing:
            return True
    return False


@lru_cache(maxsize=None)
def is_pydantic_model(klass: typing.Type[ModelProtocol]) -> bool:
    if pydantic is not None:
        return issubclass(klass, v1.BaseModel) or issubclass(klass, pydantic.BaseModel)
    return False


@lru_cache(maxsize=None)
def is_faust_record(klass: typing.Type[ModelProtocol]) -> bool:
    if faust is not None:
        return issubclass(klass, faust.Record)
    return False


def is_union(a_type: typing.Type) -> bool:
    """
    Given a python type, return True if is typing.Union, otherwise False

    Arguments:
        a_type (typing.Any): python type

    Returns:
        bool
    """
    if UnionType is None:
        return (
            isinstance(a_type, typing._GenericAlias) and a_type.__origin__ is typing.Union  # type: ignore
        )
    return (
        isinstance(a_type, typing._GenericAlias) and a_type.__origin__ is typing.Union or isinstance(a_type, UnionType)  # type: ignore
    )


def is_self_referenced(a_type: typing.Type, parent: typing.Type) -> bool:
    """
    Given a python type, return True if is self referenced, meaning
    that is instance of typing.ForwardRef, otherwise False

    Arguments:
        a_type (typing.Any): python type
        parent (typing.Any) python type

    Returns:
        bool

    Example:
        class User(...)
            a_type_with_type: typing.Type["User"]] = None
            a_type: "User" = None

        is_self_referenced(a_type) # True
    """
    return (
        isinstance(a_type, typing._GenericAlias)  # type: ignore
        and a_type.__args__
        and isinstance(a_type.__args__[0], typing.ForwardRef)
    ) or a_type == parent


def is_annotated(a_type: typing.Type) -> bool:
    """
    Given a python type, return True if is typing.Annotated, otherwise False

    Arguments:
        a_type (typing.Any): python type

    Returns:
        bool
    """
    return _is_typing_name(get_origin(a_type), name="Annotated")


def rebuild_annotation(a_type: typing.Type, field_info: FieldInfo) -> typing.Type:
    return Annotated[a_type, field_info]  # type: ignore[return-value]


def standardize_custom_type(
    *,
    field_name: str,
    value: typing.Any,
    model: ModelProtocol,
    base_class: typing.Type[ModelProtocol],
    include_type: bool = True,
) -> typing.Any:
    if isinstance(value, dict):
        return {
            k: standardize_custom_type(
                field_name=field_name, value=v, model=model, base_class=base_class, include_type=include_type
            )
            for k, v in value.items()
        }
    elif isinstance(value, list):
        return [
            standardize_custom_type(
                field_name=field_name, value=v, model=model, base_class=base_class, include_type=include_type
            )
            for v in value
        ]
    elif isinstance(value, tuple):
        return tuple(
            standardize_custom_type(
                field_name=field_name, value=v, model=model, base_class=base_class, include_type=include_type
            )
            for v in value
        )
    elif isinstance(value, enum.Enum):
        return value.value
    elif isinstance(value, base_class):
        if is_faust_record(type(value)):  # type: ignore[arg-type]
            # we need to do a trick because we can not overrride asdict from faust..
            # once the function interface is introduced we can remove this check
            asdict = value.standardize_type(include_type=include_type)  # type: ignore
        else:
            asdict = value.asdict()

        annotations = model.__annotations__
        # This is a hack to get the annotations from the parent class
        # https://github.com/marcosschroh/dataclasses-avroschema/issues/800
        if model.__class__.mro()[1] != base_class:
            annotations.update(typing.get_type_hints(model.__class__))

        if is_union(annotations[field_name]) and include_type:
            asdict["-type"] = value.get_fullname()
        return asdict

    return value


@dataclasses.dataclass
class SchemaMetadata:
    schema_name: typing.Optional[str] = None
    schema_doc: typing.Union[bool, str] = True
    namespace: typing.Optional[str] = None
    aliases: typing.Optional[typing.List[str]] = None
    alias_nested_items: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    dacite_config: typing.Optional[JsonDict] = None
    field_order: typing.Optional[typing.List[str]] = None
    exclude: typing.List[str] = dataclasses.field(default_factory=list)
    convert_literal_to_enum: bool = False

    @classmethod
    def create(cls: typing.Type["SchemaMetadata"], klass: type) -> "SchemaMetadata":
        return cls(
            schema_name=getattr(klass, "schema_name", None),
            schema_doc=getattr(klass, "schema_doc", True),
            namespace=getattr(klass, "namespace", None),
            aliases=getattr(klass, "aliases", None),
            alias_nested_items=getattr(klass, "alias_nested_items", {}),
            dacite_config=getattr(klass, "dacite_config", None),
            field_order=getattr(klass, "field_order", None),
            exclude=getattr(klass, "exclude", []),
            convert_literal_to_enum=getattr(klass, "convert_literal_to_enum", False),
        )

    def get_alias_nested_items(self, name: str) -> typing.Optional[str]:
        return self.alias_nested_items.get(name)


@dataclasses.dataclass
class FieldMetadata:
    default: typing.Any
    aliases: typing.List[str] = dataclasses.field(default_factory=list)
    doc: typing.Optional[str] = None
    namespace: typing.Optional[str] = None
    schema_name: typing.Optional[str] = None

    @classmethod
    def create(cls: typing.Type["FieldMetadata"], klass: typing.Optional[type]) -> "FieldMetadata":
        return cls(
            aliases=getattr(klass, "aliases", []),
            doc=getattr(klass, "doc", None),
            namespace=getattr(klass, "namespace", None),
            default=getattr(klass, "default", dataclasses.MISSING),
            schema_name=getattr(klass, "schema_name", None),
        )

    def to_dict(self) -> typing.Dict[str, typing.Union[typing.List[str], str]]:
        dict_repr = {key: value for key, value in vars(self).items() if value and key != "default"}

        if self.default is not dataclasses.MISSING:
            dict_repr["default"] = self.default

        return dict_repr


class UserDefinedType(typing.NamedTuple):
    name: str
    model: typing.Type["AvroModel"]


epoch: datetime = datetime(1970, 1, 1, tzinfo=timezone.utc)
epoch_naive: datetime = datetime(1970, 1, 1)



import typing
from dataclasses import MISSING, dataclass

from dataclasses_avroschema.model_generator.lang.python import templates
from dataclasses_avroschema.model_generator.lang.python.base import BaseGenerator, FieldRepresentation
from dataclasses_avroschema.types import JsonDict


@dataclass
class PydanticFieldRepresentation(FieldRepresentation):
    def get_field_metadata(self) -> typing.Optional[str]:
        metadata = {k: v for k, v in self.metadata.items() if k != "doc"}
        if metadata:
            return f"metadata={metadata}"
        return None

    def add_field_properties(self, default_repr: str) -> str:
        field_metadata_repr = self.get_field_metadata()

        if "doc" in self.metadata:
            doc = self.metadata["doc"]
            description = f'description="{doc}"'

            if field_metadata_repr:
                field_metadata_repr = f"{field_metadata_repr}, {description}"
            else:
                field_metadata_repr = description

        if field_metadata_repr or isinstance(self.default, (dict, list)):
            dataclass_field_properties = [field_metadata_repr]

            if isinstance(self.default, (dict, list)):
                if self.default:
                    dataclass_prop = f"default_factory=lambda: {default_repr}"
                else:
                    dataclass_prop = f"default_factory={default_repr}"

                dataclass_field_properties.append(dataclass_prop)
            else:
                if self.default is not MISSING:
                    dataclass_field_properties.append(f"default={default_repr}")

            default_repr = self.render_dataclass_field(
                properties=", ".join([prop for prop in dataclass_field_properties if prop])
            )

        return default_repr


@dataclass
class PydanticModelGenerator(BaseGenerator):
    def __post_init__(self) -> None:
        super().__post_init__()

        self.base_class = "pydantic.BaseModel"
        self.imports_dict = {
            "dataclass_field": "from pydantic import Field",
        }

        # Templates
        self.field_template = templates.pydantic_field_template
        self.field_representation_class = PydanticFieldRepresentation

    def _resolve_type_from_metadata(self, *, field: JsonDict) -> typing.Optional[str]:
        """
        Check if the language type must be replaced with any extra class which
        was specified in the field metadata. This method should be only called
        after the native type was resolved properly.

        An example of this if when a pydantic field was used in the model:

        class MyModel(AvroBaseModel):
            email: pydantic.EmailStr

        then the email field is represented as:

        {"name": "email", "type": {"type": "string", "pydantic-class": "EmailStr"}}

        For now we only recognize the attribute `pydantic-class` but in the future
        new way might be added, for example: `java-class`.

        """
        pydantic_class = field.get("pydantic-class")

        if pydantic_class is not None:
            return f"pydantic.{pydantic_class}"
        return None

    def add_class_imports(self) -> None:
        self.imports.add("import pydantic")



import dataclasses
import typing
from collections import OrderedDict
from typing import Literal, overload

from .types import JsonDict, SerializationType

if typing.TYPE_CHECKING:
    from .fields.base import Field  # pragma: no cover
    from .utils import UserDefinedType  # pragma: no cover

CT = typing.TypeVar("CT", bound="ModelProtocol")
# This means that when we have List[Protocol] can be use when we have List[ParserProtocol]
CP = typing.TypeVar("CP", bound="ParserProtocol", covariant=True)


class ParserProtocol(typing.Protocol[CP]):
    """
    Parse python dataclasses to represent it as an avro schema.

    Each classes field is converted into an AvroField in order to
    be represented as an avro type.
    """

    def __init__(self, type: typing.Type["ModelProtocol"], parent: typing.Type["ModelProtocol"]): ...

    def generate_dataclass(self) -> typing.Type: ...

    def parse_fields(self, exclude: typing.List) -> typing.List["Field"]: ...

    def get_fields_map(self) -> typing.Dict[str, "Field"]: ...

    def get_schema_name(self) -> str: ...

    def generate_documentation(self) -> typing.Optional[str]: ...

    def get_rendered_fields(self) -> typing.List[OrderedDict]: ...

    def render(self) -> OrderedDict: ...


class ModelProtocol(typing.Protocol[CT]):
    # _parser: typing.Optional[ParserProtocol] = None
    _parent: typing.Optional[CT] = None
    _user_defined_types: typing.Set["UserDefinedType"] = set()
    _rendered_schema: OrderedDict = dataclasses.field(default_factory=OrderedDict)

    @classmethod
    def get_fullname(cls: typing.Type[CT]) -> str: ...

    @classmethod
    def generate_schema(
        cls: typing.Type[CT], schema_type: SerializationType = "avro"
    ) -> typing.Optional[OrderedDict]: ...

    @classmethod
    def _get_serialization_context(cls: typing.Type[CT]) -> JsonDict: ...

    @classmethod
    def _generate_parser(cls: typing.Type[CT]) -> ParserProtocol: ...

    @classmethod
    def avro_schema(cls: typing.Type[CT], case_type: typing.Optional[str] = None, **kwargs) -> str: ...

    @classmethod
    def avro_schema_to_python(
        cls: typing.Type[CT],
        parent: typing.Optional[CT] = None,
        case_type: typing.Optional[str] = None,
    ) -> typing.Dict[str, typing.Any]: ...

    @classmethod
    def get_fields(cls: typing.Type[CT]) -> typing.List["Field"]: ...

    @classmethod
    def _reset_parser(cls: typing.Type[CT]) -> None: ...

    @classmethod
    @overload
    def deserialize(
        cls: typing.Type[CT],
        data: bytes,
        serialization_type: SerializationType = "avro",
        create_instance: Literal[True] = ...,
        writer_schema: typing.Optional[typing.Union[JsonDict, typing.Type[CT]]] = None,
    ) -> CT: ...

    @classmethod
    @overload
    def deserialize(
        cls: typing.Type[CT],
        data: bytes,
        serialization_type: SerializationType = "avro",
        create_instance: Literal[False] = ...,
        writer_schema: typing.Optional[typing.Union[JsonDict, typing.Type[CT]]] = None,
    ) -> JsonDict: ...

    @classmethod
    @overload
    def deserialize(
        cls: typing.Type[CT],
        data: bytes,
        serialization_type: SerializationType = "avro",
        create_instance: bool = ...,
        writer_schema: typing.Optional[typing.Union[JsonDict, typing.Type[CT]]] = None,
    ) -> typing.Union[JsonDict, CT]: ...

    @classmethod
    def deserialize_to_python(  # This can be used straight with a pydantic dataclass to bypass dacite
        cls: typing.Type[CT],
        data: bytes,
        serialization_type: SerializationType = "avro",
        writer_schema: typing.Union[JsonDict, typing.Type[CT], None] = None,
    ) -> dict: ...

    @classmethod
    def parse_obj(cls: typing.Type[CT], data: typing.Dict) -> CT: ...

    @classmethod
    def fake(cls: typing.Type[CT], **data: typing.Any) -> CT: ...

    def asdict(self: CT) -> JsonDict: ...

    def serialize(self: CT, serialization_type: SerializationType = "avro") -> bytes: ...

    # def validate(self: CT) -> bool: ...

    def to_dict(self: CT) -> JsonDict: ...

    def to_json(self: CT, **kwargs: typing.Any) -> str: ...



import typing

from fastavro.validation import validate

from dataclasses_avroschema import AvroModel, serialization
from dataclasses_avroschema.types import JsonDict
from dataclasses_avroschema.utils import standardize_custom_type

from .parser import FaustParser

try:
    from faust import Record  # pragma: no cover
except ImportError as ex:  # pragma: no cover
    raise Exception("faust-streaming must be installed in order to use AvroRecord") from ex  # pragma: no cover


class AvroRecord(Record, AvroModel):  # type: ignore
    def validate_avro(self) -> bool:
        """
        Validate that instance matches the avro schema
        """
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    def standardize_type(self, include_type: bool = True) -> typing.Any:
        """
        Standardization factory that converts data according to the
        user-defined pydantic json_encoders prior to passing values
        to the standard type conversion factory
        """
        return {
            field_name: standardize_custom_type(
                field_name=field_name, value=value, model=self, base_class=AvroRecord, include_type=include_type
            )
            for field_name, value in self.asdict().items()
        }

    def serialize(self, serialization_type: serialization.SerializationType = "avro") -> bytes:
        """
        Overrides the base AvroModel's serialize method to inject this
        class's standardization factory method
        """
        schema = self.avro_schema_to_python()
        data = self.standardize_type()

        return serialization.serialize(
            data,
            schema,
            serialization_type=serialization_type,
        )

    def to_dict(self) -> JsonDict:
        return self.standardize_type(include_type=False)

    @classmethod
    def _generate_parser(cls: typing.Type["AvroRecord"]) -> FaustParser:
        return FaustParser(type=cls, parent=cls._parent or cls)



import datetime
import decimal
import enum
import io
import typing
import uuid

import fastavro

from .protocol import ModelProtocol
from .types import JsonDict, SerializationType

DATETIME_STR_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
DATE_STR_FORMAT = "%Y-%m-%d"
TIME_STR_FORMAT = "%H:%M:%S"

AVRO = "avro"
AVRO_JSON = "avro-json"

decimal_context = decimal.Context()


def serialize(payload: JsonDict, schema: typing.Dict, serialization_type: SerializationType = "avro") -> bytes:
    """
    Serialize a payload into avro using `fastavro` as backend

    Attributes:
        payload typing.Dict[str, Any]: The payload to serialize
        serialization_type SerializationType: `avro` or `avro-json`

    Returns:
        bytes encoded in avro format

    !!! Example
        ```python
        from dataclasses_avroschema import serialize


        payload = {'event': 'Hello world'}
        schema = {
            'type': 'record',
            'name': 'MyRecord',
            'fields': [
                {'name': 'event', 'type': 'string', 'default': 'Hello World'}
            ]
        }

        event = serialize(payload=payload, schema=schema)
        assert event == b'\x16Hello world'
        ```
    """
    if serialization_type == AVRO:
        file_like_output: typing.Union[io.BytesIO, io.StringIO] = io.BytesIO()

        fastavro.schemaless_writer(file_like_output, schema, payload)

        value = file_like_output.getvalue()
    elif serialization_type == AVRO_JSON:
        file_like_output = io.StringIO()
        fastavro.json_writer(file_like_output, schema, [payload])
        value = file_like_output.getvalue().encode("utf-8")
    else:
        raise ValueError(f"Serialization type should be `avro` or `avro-json`, not {serialization_type}")

    file_like_output.flush()

    return value  # type: ignore


def deserialize(
    *,
    data: bytes,
    schema: JsonDict,
    serialization_type: SerializationType = "avro",  # ADd enum
    context: typing.Optional[JsonDict] = None,
    writer_schema: typing.Optional[JsonDict] = None,
) -> JsonDict:
    """
    Deserialize an binary `event` into a python Dict using `fastavro` as backend

    Attributes:
        data bytes: The event to deserialize
        schema: Dict[str, Any]: The schema to use for the deserialization
        serialization_type SerializationType: `avro` or `avro-json`
        context Dict[str, Any] | None: Optional extra context to use.
            Usually in includes an entry with all the extra models defined
            by the end user by name. Example AvroModel.__name__: AvroModel
        writer_schema Dict[str, Any] | None: The schema that was used to write
            the event. If it is not provided it is assumed that the event was
            written with the `schema` provided

    Returns:
        The object dezerialized python Dict

    !!! Example
        ```python
        from dataclasses_avroschema import deserialize


        event = b'\x16Hello world'
        schema = {
            'type': 'record',
            'name': 'MyRecord',
            'fields': [
                {'name': 'event', 'type': 'string', 'default': 'Hello World'}
            ]
        }

        payload = deserialize(data=event, schema=schema)
        assert payload == {'event': 'Hello world'}
        ```
    """

    if serialization_type == AVRO:
        input_stream: typing.Union[io.BytesIO, io.StringIO] = io.BytesIO(data)

        payload = fastavro.schemaless_reader(
            input_stream,
            writer_schema=writer_schema or schema,
            reader_schema=schema,
            return_record_name=True,
            return_record_name_override=True,
        )

    elif serialization_type == AVRO_JSON:
        input_stream = io.StringIO(data.decode())
        # This is an iterator, but not a container
        records = fastavro.json_reader(input_stream, schema)
        # records can have multiple payloads, but in this case we return the first one
        payload = list(records)[0]
    else:
        raise ValueError(f"Serialization type should be `avro` or `avro-json`, not {serialization_type}")

    input_stream.flush()

    if context is not None:
        return deserialize_from_context(data=payload, context=context)  # type: ignore
    return payload  # type: ignore


def deserialize_from_context(*, data: JsonDict, context: JsonDict) -> JsonDict:
    """
    This function tries to convert cast all the cases that have
    `unions` with a Tuple format (AvroType, payload), for example
    (Car, {"brand": "fiat"}). This cases happens when there are
    avro unions of records which are similar or equals among them
    """
    cleaned_data = {}
    for field_name, field_value in data.items():
        if isinstance(field_value, dict):
            field_value = deserialize_from_context(data=field_value, context=context)
        elif isinstance(field_value, tuple) and len(field_value) == 2:
            field_value = sanitize_union(union=field_value, context=context)

        cleaned_data[field_name] = field_value

    return cleaned_data


def sanitize_union(*, union: typing.Tuple, context: JsonDict) -> typing.Optional[ModelProtocol]:
    # the first value is the model/record name and the second is its payload
    model_name, model_value = union
    if isinstance(model_value, dict):
        # it can be a dict again so we need to sanitize
        model_value = deserialize_from_context(data=model_value, context=context)

    model_name = model_name.split(".")[-1]
    avro_model: typing.Optional[ModelProtocol] = context.get(model_name)

    if avro_model is not None:
        return avro_model.parse_obj(model_value)

    return model_value


def datetime_to_str(value: datetime.datetime) -> str:
    return value.strftime(DATETIME_STR_FORMAT)


def date_to_str(value: datetime.date) -> str:
    return value.strftime(DATE_STR_FORMAT)


def time_to_str(value: datetime.time) -> str:
    return value.strftime(TIME_STR_FORMAT)


def decimal_to_str(value: decimal.Decimal, precision: int, scale: int = 0) -> str:
    value_bytes = prepare_bytes_decimal(value, precision, scale)
    return r"\u" + value_bytes.hex()


def string_to_decimal(*, value: str, schema: JsonDict) -> decimal.Decimal:
    # first remove the Unicode character
    value = value.replace(r"\u", "")

    # then concert to bytes
    decimal_bytes = bytes.fromhex(value)

    # finally get the decimal.Decimal
    scale = schema.get("scale", 0)
    precision = schema["precision"]
    unscaled_datum = int.from_bytes(decimal_bytes, byteorder="big", signed=True)
    decimal_context.prec = precision

    return decimal_context.create_decimal(unscaled_datum).scaleb(-scale, decimal_context)


# This is an almost complete copy of fastavro's _logical_writers_py.prepare_bytes_decimal
# the only tweak is to pass in scale/precision directly instead of a schema
# This is needed to properly serialize a default decimal.Decimal into the avro schema
def prepare_bytes_decimal(data: decimal.Decimal, precision: int, scale: int = 0) -> bytes:
    """Convert decimal.Decimal to bytes"""
    sign, digits, exp = data.as_tuple()

    if len(digits) > precision:
        raise ValueError("The decimal precision is bigger than allowed by schema")

    delta = int(exp) + scale

    if delta < 0:
        raise ValueError("Scale provided in schema does not match the decimal")

    unscaled_datum = 0
    for digit in digits:
        unscaled_datum = (unscaled_datum * 10) + digit

    unscaled_datum = 10**delta * unscaled_datum

    bytes_req = (unscaled_datum.bit_length() + 8) // 8

    if sign:
        unscaled_datum = -unscaled_datum

    return unscaled_datum.to_bytes(bytes_req, byteorder="big", signed=True)


def serialize_value(*, value: typing.Any) -> typing.Any:
    from .main import AvroModel

    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, datetime.datetime):
        value = datetime_to_str(value)
    elif isinstance(value, datetime.date):
        value = date_to_str(value)
    elif isinstance(value, datetime.time):
        value = time_to_str(value)
    elif isinstance(value, datetime.timedelta):
        value = value.total_seconds()
    elif isinstance(value, (uuid.UUID, decimal.Decimal)):
        value = str(value)
    elif isinstance(value, dict):
        value = to_json(value)
    elif isinstance(value, enum.Enum):
        value = value.value
    elif isinstance(value, (list, tuple)):
        value = type(value)(serialize_value(value=item) for item in value)
    elif isinstance(value, AvroModel):
        value = to_json(value.asdict())

    return value


def to_json(data: JsonDict) -> JsonDict:
    json_data = {}

    for field, value in data.items():
        json_data[field] = serialize_value(value=value)

    return json_data



import copy
import dataclasses
import json
import keyword
import typing
from abc import abstractmethod
from dataclasses import dataclass, field
from string import Template

import casefy

from dataclasses_avroschema import serialization
from dataclasses_avroschema.fields import field_utils
from dataclasses_avroschema.types import AvroTypeRepr, JsonDict

from . import (
    avro_to_python_utils,
    templates,
)


@dataclasses.dataclass
class FieldRepresentation:
    name: str
    avro_type: typing.Any
    type_hint: str
    default: typing.Any
    template: Template
    metadata: JsonDict = dataclasses.field(default_factory=dict)
    children: typing.List["FieldRepresentation"] = dataclasses.field(default_factory=list)
    inner_name: str = ""

    def __post_init__(self) -> None:
        if isinstance(self.avro_type, list):
            self.avro_type = self.avro_type[0]

        if isinstance(self.avro_type, dict):
            inner_type = self.avro_type["type"]
            self.inner_name = self.avro_type.get("name", "")

            if self.children:
                self.child_field = self.children[0]
                self.avro_type = self.child_field.avro_type
                self.type_hint = self.type_hint or self.child_field.type_hint
                self.default = self.default if self.default is not dataclasses.MISSING else self.child_field.default

            # Check that internal field name is different to the original name
            # for example: {"name": "age", "type": { "type": "array", "items": "string", "name": "my_age" }}
            # in this case my_age is not in age which means that this must be reflected in the model
            # This does not applu to `enums` and `records`
            if self.inner_name not in self.name and inner_type not in (
                field_utils.ENUM,
                field_utils.RECORD,
            ):
                self.metadata["inner_name"] = self.inner_name

        # update all metadata based on all children
        for child in self.children:
            self.metadata.update(child.metadata)

    def render(self, type_hint_clashes: dict[str, str]) -> str:
        """
        Convert a field to string
        """
        type_hint = type_hint_clashes.get(self.type_hint, self.type_hint)
        result = templates.field_type_template.safe_substitute(name=self.name, type=type_hint)

        # optional field attribute
        default_generated = self.get_field_default()

        if (
            default_generated
            not in (
                dataclasses.MISSING,
                "",
            )
            and self.avro_type != field_utils.DECIMAL
        ):
            result += templates.field_default_template.safe_substitute(default=default_generated)

        return result

    def get_field_metadata(self) -> typing.Optional[str]:
        if self.metadata:
            return f"metadata={self.metadata}"
        return None

    def add_field_properties(self, default_repr: str) -> str:
        field_metadata_repr = self.get_field_metadata()

        if field_metadata_repr or isinstance(self.default, (dict, list)):
            dataclass_field_properties = [field_metadata_repr]

            if isinstance(self.default, (dict, list)):
                if self.default:
                    dataclass_prop = f"default_factory=lambda: {default_repr}"
                else:
                    dataclass_prop = f"default_factory={default_repr}"

                dataclass_field_properties.append(dataclass_prop)
            else:
                if self.default is not dataclasses.MISSING:
                    dataclass_field_properties.append(f"default={default_repr}")

            default_repr = self.render_dataclass_field(
                properties=", ".join([prop for prop in dataclass_field_properties if prop])
            )

        return default_repr

    def get_field_default(self) -> str:
        """
        Returns the default value according to the field type

        TODO: docstrings

        Example:
            If the default is "bond" the method should return '"bond"' so the double quotes
            won't be scaped during the field render
        """
        default_repr = ""

        if self.default is dataclasses.MISSING:
            ...
        elif self.avro_type in (
            field_utils.STRING,
            field_utils.UUID,
        ):
            default_repr = f'"{self.default}"'
        elif self.avro_type in (
            field_utils.BYTES,
            field_utils.FIXED,
        ):
            default_repr = f'b"{self.default}"'
        elif self.avro_type == field_utils.ENUM:
            enum_name = self.inner_name or self.type_hint
            default_repr = f"{casefy.pascalcase(enum_name)}.{casefy.uppercase(self.default)}"
        elif isinstance(self.default, (dict, list)):
            # Then is can be a regular dict as default or a record
            if self.default:
                if self.avro_type not in field_utils.AVRO_TYPES or self.avro_type == field_utils.RECORD:
                    # Try to get the last part in case that the type is namespaced `types.bus_type.Bus`
                    field_type = self.avro_type.split(".")[-1]

                    if field_type == field_utils.RECORD:
                        # Then it is a record defined by the first time with a default value
                        field_type = self.type_hint

                    default_repr = templates.instance_template.safe_substitute(
                        type=field_type, properties=f"**{self.default}"
                    )
                else:
                    default_repr = str(self.default)
            else:
                default_repr = self.default.__class__.__name__
        elif self.avro_type in avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON:
            func = avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON[self.avro_type]
            python_type = func(self.default)
            template_func = avro_to_python_utils.LOGICAL_TYPE_TEMPLATES[self.avro_type]
            default_repr = template_func(python_type)
        elif self.avro_type == field_utils.NULL:
            default_repr = "None"
        elif self.avro_type == self.type_hint:
            # this is when a Custom type is been reused which is probably an Enum
            default_repr = f"{casefy.pascalcase(self.type_hint)}.{casefy.uppercase(self.default)}"
        else:
            default_repr = str(self.default)

        default_repr = self.add_field_properties(default_repr)

        return default_repr

    def render_dataclass_field(self, properties: str) -> str:
        return self.template.safe_substitute(properties=properties)


@dataclass
class ClassRepresentation:
    name: str
    schema: JsonDict
    base_class: str
    template: Template = templates.class_template
    metaclass_template: Template = templates.metaclass_template
    field_identation: str = "\n    "
    metadata_field_templates: typing.Dict[str, Template] = field(
        default_factory=lambda: {
            "namespace": templates.metaclass_field_template,
            "doc": templates.metaclass_field_template,
            "aliases": templates.metaclass_alias_field_template,
            "original_schema": templates.metaclass_schema_field_template,
            "default": templates.metaclass_field_template,
            "schema_name": templates.metaclass_field_template,
        }
    )
    metadata_fields_mapper: typing.Dict[str, str] = field(
        default_factory=lambda: {
            # doc is not included because it is rendered as docstrings
            "namespace": "namespace",
            "aliases": "aliases",
            "default": "default",
        }
    )
    # Boolean to indicate whether original_schema field containing the original schema string should be generated in
    # Meta class of all generated objects
    include_original_schema: bool = False

    @abstractmethod
    def render(self, type_hint_clashes: dict[str, str]) -> str: ...

    def render_metaclass(
        self,
        *,
        schema: JsonDict,
        field_order: typing.Optional[typing.List[str]] = None,
        decorator: str = "",
        add_schema_name: bool = False,
    ) -> typing.Optional[str]:
        """
        Render Class Meta that contains the schema matadata
        """
        metadata = [
            self.metadata_field_templates[meta_avro_field].safe_substitute(name=meta_field, value=value)
            for meta_avro_field, meta_field in self.metadata_fields_mapper.items()
            if (value := schema.get(meta_avro_field))
        ]

        # Parse the entire schema string into the Meta class field "original_schema" for each class.
        # If self.include_original_schema is set to True.
        if self.include_original_schema:
            metadata.append(self._add_schema_to_metaclass(self.metadata_field_templates["original_schema"], schema))

        if field_order is not None:
            metadata.append(
                templates.metaclass_alias_field_template.safe_substitute(
                    name="field_order",
                    value=field_order,
                )
            )

        if add_schema_name:
            metadata.append(
                self.metadata_field_templates["schema_name"].safe_substitute(name="schema_name", value=schema["name"])
            )

        properties = self.field_identation.join(metadata)

        if properties:
            # some formating to remove identation at the end of the Class Meta to make it more compatible with black
            return self.field_identation.join(
                [
                    line
                    for line in self.metaclass_template.safe_substitute(
                        properties=properties, decorator=decorator
                    ).split("\n")
                ]
            ).rstrip(self.field_identation)
        return None

    def render_docstring(self, *, docstring: typing.Optional[str]) -> str:
        """
        Render the module with the classes generated from the schema
        """
        if not docstring:
            return ""

        indented = self.field_identation + self.field_identation.join(docstring.splitlines())

        return f'{self.field_identation}"""{indented}{self.field_identation}"""'

    @staticmethod
    def _add_schema_to_metaclass(schema_template: Template, schema: JsonDict) -> str:
        """
        Parses provided schema to ModelGenerator.render() to a Meta field string by using
        metaclass_schema_field_template.
        """
        return schema_template.safe_substitute(name="original_schema", schema=json.dumps(schema))


@dataclass
class EnumRepresentation(ClassRepresentation):
    symbols: JsonDict = field(default_factory=dict)

    def render(self, type_hint_clashes: dict[str, str]) -> str:
        symbols_repr = self.render_symbols()

        docstring = self.render_docstring(docstring=self.schema.get("doc"))
        enum_repr = templates.enum_template.safe_substitute(name=self.name, symbols=symbols_repr, docstring=docstring)
        add_schema_name = self.name != self.schema["name"]
        metaclass = self.render_metaclass(
            schema=self.schema,
            decorator=templates.METACLASS_DECORATOR,
            add_schema_name=add_schema_name,
        )

        if metaclass:
            enum_repr += metaclass

        if self.name in type_hint_clashes:
            clashed_typing = templates.metaclass_alias_field_template.safe_substitute(
                name=type_hint_clashes[self.name], value=self.name
            )
            clashed_typing = f"\n{clashed_typing}\n"
            enum_repr += clashed_typing

        return f"{enum_repr}\n"

    def render_symbols(self) -> str:
        return self.field_identation.join(
            [
                templates.enum_symbol_template.safe_substitute(key=key, value=f'"{value}"')
                for key, value in self.symbols.items()
            ]
        )


@dataclass
class ModelRepresentation(ClassRepresentation):
    fields_representation: typing.List[FieldRepresentation] = field(default_factory=list)
    # represent the decorator to add in the base class
    decorator: typing.Optional[str] = None

    def render(self, type_hint_clashes: dict[str, str]) -> str:
        fields_representation_copy = copy.copy(self.fields_representation)
        self.fields_representation.sort(key=lambda field: field.default is not dataclasses.MISSING)

        field_order = None
        if fields_representation_copy != self.fields_representation:
            field_order = [field.name for field in fields_representation_copy]

        rendered_fields_string = (
            self.field_identation.join(
                field.render(type_hint_clashes=type_hint_clashes) for field in self.fields_representation
            )
            if len(self.fields_representation) > 0
            else "pass"
        )
        docstring = self.render_docstring(docstring=self.schema.get("doc"))

        rendered_class = self.template.safe_substitute(
            name=self.name,
            decorator=self.decorator or "",
            base_class=self.base_class,
            fields=rendered_fields_string,
            docstring=docstring,
        )

        add_schema_name = self.name != self.schema["name"]
        class_metadata = self.render_metaclass(
            schema=self.schema, field_order=field_order, add_schema_name=add_schema_name
        )
        if class_metadata is not None:
            rendered_class += class_metadata

        if self.name in type_hint_clashes:
            clashed_typing = templates.metaclass_alias_field_template.safe_substitute(
                name=type_hint_clashes[self.name], value=self.name
            )
            clashed_typing = f"\n{clashed_typing}\n"
            rendered_class += clashed_typing

        return f"{rendered_class}\n"


@dataclass
class BaseGenerator:
    field_identation: str = "\n    "
    imports: typing.Set[str] = field(default_factory=set)
    imports_dict: typing.Dict[str, str] = field(default_factory=dict)
    field_representation_class: typing.Type[FieldRepresentation] = FieldRepresentation

    # extras is used fot extra code that is generated, for example Enums
    extras: typing.List[str] = field(default_factory=list)
    classes_representation: typing.List[ClassRepresentation] = field(default_factory=list)

    field_template: Template = field(init=False)
    avro_type_to_lang: typing.Dict[str, str] = field(init=False)
    logical_types_imports: typing.Dict[str, str] = field(init=False)
    base_class_decorator: typing.Optional[str] = None
    # Boolean to indicate whether original_schema field containing the original schema string should be generated in
    # Meta class of all generated objects
    include_original_schema: bool = False
    base_class: str = field(init=False)
    type_hint_clashes: dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.avro_type_to_lang = avro_to_python_utils.AVRO_TYPE_TO_PYTHON
        self.logical_types_imports = avro_to_python_utils.LOGICAL_TYPES_IMPORTS

    def add_class_imports(self) -> None: ...

    def _resolve_type_from_metadata(self, *, field: JsonDict) -> typing.Optional[str]: ...

    def render(self, schemas: typing.List[JsonDict]) -> str:
        for schema in schemas:
            if "fields" in schema:
                self.render_class(schema=schema)
            else:
                # If is not a Record then it is an Enum
                self.render_field(field=schema, model_name="", parent_field_name="")

        result = [
            class_repr.render(type_hint_clashes=self.type_hint_clashes) for class_repr in self.classes_representation
        ]
        classes = "".join(result)
        imports = self.render_imports()
        extras = self.render_extras()

        return templates.module_template.safe_substitute(
            classes=classes,
            imports=imports,
            extras=extras,
        ).lstrip("\n")

    def render_imports(self) -> str:
        """
        Render the imports needed for the python classes.
        """
        # sort the imports
        list_imports = list(self.imports)
        list_imports.sort()

        imports = "\n".join([imp for imp in list_imports])
        return templates.imports_template.safe_substitute(imports=imports)

    def render_extras(self) -> str:
        return "".join([extra for extra in self.extras])

    def render_class(self, *, schema: JsonDict, parent_field_name: typing.Optional[str] = None) -> ClassRepresentation:
        """
        Render the class generated from the schema
        """
        model_name: str = casefy.pascalcase(schema["name"])

        record_fields: typing.List[JsonDict] = schema["fields"]
        self.add_class_imports()

        # Sort the fields according whether it has a default value
        fields_representation: typing.List[FieldRepresentation] = [
            self.render_field(field=field, model_name=model_name, parent_field_name=parent_field_name or field["name"])
            for field in record_fields
        ]

        class_representation = ModelRepresentation(
            name=model_name,
            schema=schema,
            fields_representation=fields_representation,
            base_class=self.base_class,
            decorator=self.base_class_decorator,
            include_original_schema=self.include_original_schema,
        )
        self.classes_representation.append(class_representation)

        return class_representation

    def render_field(self, field: JsonDict, model_name: str, parent_field_name: str) -> FieldRepresentation:
        """
        Render an avro field.

        1. If the field is a native one, we just render it.
        2. If the field is a record, array, map, fixed, enum
            we need to render the field again (recursive)
        3. If the field is a LogicalType, it may not have the
            the `name` property and the type is a `native` one
        """
        name = self.generate_field_name(field)

        avro_type: AvroTypeRepr = field["type"]
        default = field.get("default", dataclasses.MISSING)
        field_metadata = self.get_field_metadata(field)
        children = None

        if self.is_logical_type(field=field):
            # override the type so it can be use to get the default value in case that is needed
            avro_type = field.get("logicalType") or field["type"]["logicalType"]
            type_hint = self.parse_logical_type(field=field)
        elif isinstance(avro_type, dict):
            type_hint = ""
            children = [
                self.render_field(field=avro_type, model_name=model_name, parent_field_name=name or parent_field_name)
            ]
        elif isinstance(avro_type, list):
            type_hint, children = self.parse_union(field_types=avro_type, model_name=model_name, parent_field_name=name)
        elif avro_type == field_utils.ARRAY:
            type_hint, children = self.parse_array(
                field=field, model_name=model_name, parent_field_name=name or parent_field_name
            )
        elif avro_type == field_utils.MAP:
            type_hint, children = self.parse_map(
                field=field, model_name=model_name, parent_field_name=name or parent_field_name
            )
        elif avro_type == field_utils.ENUM:
            type_hint = self.parse_enum(field=field)
            # We must set the default Enums type level default to dataclasses.MISSING
            # as it is set in the Meta class.
            # Check https://github.com/marcosschroh/dataclasses-avroschema/issues/665
            default = dataclasses.MISSING
        elif avro_type == field_utils.FIXED:
            type_hint = self.parse_fixed(field=field)
        elif avro_type == field_utils.RECORD:
            type_hint = casefy.pascalcase(name)
            class_representation = self.render_class(schema=field, parent_field_name=name)

            if class_representation not in self.classes_representation:
                self.classes_representation.append(class_representation)
        else:
            # Native field
            type_hint = self.get_language_type(avro_type=avro_type, field_name=name, model_name=model_name)

            # check if the language type must be replaced with some extra class
            # specified in the field metadata only after the native type was resolved
            type_from_metadata = self._resolve_type_from_metadata(field=field)
            type_hint = type_from_metadata or type_hint

        if type_hint == parent_field_name:
            # then there is a clashes type
            type_hint = f"_{parent_field_name}"
            self.type_hint_clashes[parent_field_name] = type_hint

            if name:
                self.type_hint_clashes[name] = type_hint

        return self.field_representation_class(
            name=name,
            avro_type=avro_type,
            type_hint=type_hint,
            template=self.field_template,
            default=default,
            metadata=field_metadata,
            children=children or [],
        )

    @staticmethod
    def generate_field_name(field: JsonDict) -> str:
        field_name = field.get("name", "")
        if not field_name:
            return field_name

        if keyword.iskeyword(field_name):
            valid_identifier = f"{field_name}_"
            BaseGenerator.maybe_add_alias(field, field_name)
            return valid_identifier

        if not field_name.isidentifier():
            valid_identifier = casefy.snakecase(field_name)
            if valid_identifier != field_name:
                BaseGenerator.maybe_add_alias(field, field_name)
            return valid_identifier
        return field_name

    @staticmethod
    def maybe_add_alias(field: JsonDict, alias: str) -> None:
        aliases = field.get("aliases", [])

        if alias not in aliases:
            aliases.append(alias)
            field["aliases"] = aliases

    @staticmethod
    def is_logical_type(*, field: JsonDict) -> bool:
        if field.get("logicalType"):
            return True

        field_type = field["type"]
        return isinstance(field_type, dict) and field_type.get("logicalType") is not None

    def parse_logical_type(self, *, field: JsonDict) -> str:
        field_name = field.get("name")
        default = field.get("default")
        field = field if field_name is None or field.get("logicalType") else field["type"]
        logical_type = field["logicalType"]

        type_from_metadata = self._resolve_type_from_metadata(field=field)

        if type_from_metadata is not None:
            type_hint = type_from_metadata
        elif logical_type == field_utils.DECIMAL:
            # this is a special case for logical types
            type_hint = self.parse_decimal(field=field, default=default)
        elif logical_type not in self.logical_types_imports:
            # Then it is a custom logicalType, so we default to the native type
            type_hint = self.get_language_type(avro_type=field["type"])
        else:
            # add the logical type import
            self.imports.add(self.logical_types_imports[logical_type])
            type_hint = self.avro_type_to_lang[logical_type]

        return type_hint

    def parse_decimal(self, *, field: JsonDict, default: typing.Optional[str] = None) -> str:
        precision = field["precision"]
        scale = field["scale"]

        self.imports.add("from dataclasses_avroschema import types")
        field_repr = templates.decimal_type_template.safe_substitute(precision=precision, scale=scale)

        if default is not None:
            self.imports.add("import decimal")
            default = templates.decimal_template.safe_substitute(
                value=serialization.string_to_decimal(value=default, schema=field)
            )
            field_repr += templates.field_default_template.safe_substitute(default=default)
        return field_repr

    def parse_union(
        self, *, field_types: typing.List, model_name: str, parent_field_name: str
    ) -> typing.Tuple[str, typing.List[FieldRepresentation]]:
        """
        Parse an Avro union

        An union field is an array like ["null", "str", "int", ...]

        If the first element is `null` it means that the property is optional.
        The first item is the actual type of the field a AVRO specifies.

        Attributes:
            field_types: List of avro types
            model_name: name of the model that contains the `union`
        """

        # XXX: Maybe more useful in general
        def render_type(avro_type: typing.Union[str, dict]) -> FieldRepresentation:
            avro_type = self.type_to_defined_types(avro_type)

            return self.render_field(field=avro_type, model_name=model_name, parent_field_name=parent_field_name)

        if field_utils.NULL in field_types and len(field_types) == 2:
            # It is an optional field, we should include in the imports typing
            # and use the optional Template
            self.imports.add("import typing")
            field_type = [f for f in field_types if f != field_utils.NULL]
            field_representation = render_type(field_type[0])
            type_hint = templates.optional_template.safe_substitute(type=field_representation.type_hint)
            return type_hint, [field_representation]
        elif len(field_types) >= 2:
            # a union with more than 2 types
            self.imports.add("import typing")
            children = [render_type(t) for t in field_types]
            language_types_repr = ", ".join(child.type_hint for child in children)
            type_hint = templates.union_template.safe_substitute(type=language_types_repr)
            return type_hint, children
        else:
            # This is a weird definitions when an union can have only one type
            #  {"name": "money_available", "type": ["double"]},
            field_representation = render_type(field_types[0])
            return field_representation.type_hint, [field_representation]

    def parse_array(
        self, field: JsonDict, model_name: str, parent_field_name: str
    ) -> typing.Tuple[str, typing.List[FieldRepresentation]]:
        """
        Parse an Avro array

        The type is is specify in the `items` attribute

        Example:
            {"name": "pets", "type": {"type": "array", "items": "string", "name": "pet"}}
        """
        self.imports.add("import typing")
        avro_type = self.type_to_defined_types(field["items"])

        field_representation = self.render_field(
            field=avro_type, model_name=model_name, parent_field_name=parent_field_name
        )
        type_hint = templates.list_template.safe_substitute(type=field_representation.type_hint)

        return type_hint, [field_representation]

    def parse_map(
        self, field: JsonDict, model_name: str, parent_field_name: str
    ) -> typing.Tuple[str, typing.List[FieldRepresentation]]:
        """
        Parse an Avro map

        The type is is specify in the `values` attribute. All the keys must be `string`

        Example:
            {"name": "accounts_money", "type": {"type": "map", "values": "float", "name": "accounts_money"}},
        """
        self.imports.add("import typing")
        avro_type = self.type_to_defined_types(field["values"])

        field_representation = self.render_field(
            field=avro_type, model_name=model_name, parent_field_name=parent_field_name
        )
        type_hint = templates.dict_template.safe_substitute(type=field_representation.type_hint)

        return type_hint, [field_representation]

    def parse_fixed(self, field: JsonDict) -> str:
        self.imports.add("from dataclasses_avroschema import types")
        properties = f"size={field['size']}"

        namespace = field.get("namespace")
        aliases = field.get("aliases")

        if namespace is not None:
            properties += f', namespace="{namespace}"'

        if aliases is not None:
            properties += f", aliases={aliases}"

        return templates.fixed_template.safe_substitute(properties=properties)

    def parse_enum(self, field: JsonDict) -> str:
        """
        Parse an Avro Enum

        Avro Enums are asociated with enum.Enum types in python.
        We need a template for it in order to create the Enum
        """
        self.imports.add("import enum")
        enum_name: str = casefy.pascalcase(field["name"])

        symbols_map = {}
        symbols: typing.List[str] = field["symbols"]
        for symbol in symbols:
            key = casefy.uppercase(symbol)
            if key in symbols:
                key = symbol
            symbols_map[key] = symbol

        self.classes_representation.append(
            EnumRepresentation(
                name=enum_name,
                schema=field,
                template=templates.enum_template,
                base_class="",
                symbols=symbols_map,
            )
        )

        return enum_name

    def get_language_type(
        self,
        *,
        avro_type: str,
        field_name: typing.Optional[str] = None,
        default: typing.Optional[str] = None,
        model_name: typing.Optional[str] = None,
    ) -> str:
        if avro_type in (field_utils.INT, field_utils.FLOAT):
            self.imports.add("from dataclasses_avroschema import types")

        if avro_type == model_name:
            # it means that it is a one-to-self-relationship
            return templates.type_template.safe_substitute(type=avro_type)
        elif avro_type not in self.avro_type_to_lang:
            # it means the type points to the an already specified type so it contains the type name
            # with optional namespaces, e.g. my_namespace.users.User
            # In this case we should return the last part of the string
            return casefy.pascalcase(avro_type.split(".")[-1])
        elif default is not None:
            return str(self.avro_type_to_lang.get(avro_type, default))
        else:
            return str(self.avro_type_to_lang.get(avro_type, avro_type))

    @staticmethod
    def type_to_defined_types(avro_type: typing.Any) -> JsonDict:
        if isinstance(avro_type, (str, list)):
            avro_type = {"type": avro_type}

        return avro_type

    def get_field_metadata(self, field: JsonDict) -> JsonDict:
        keys_to_ignore = [
            "name",
            "type",
            "default",
            "pydantic-class",
            "precision",
            "scale",
        ]

        type = field["type"]

        if type == field_utils.ARRAY:
            keys_to_ignore.append("items")
        elif type == field_utils.MAP:
            keys_to_ignore.append("values")
        elif type == field_utils.FIXED:
            keys_to_ignore.extend(["namespace", "aliases", "size"])
        elif type == field_utils.ENUM:
            keys_to_ignore.extend(["symbols", "namespace", "aliases", "doc"])
        elif type == field_utils.RECORD:
            keys_to_ignore.extend(["fields", "doc", "aliases", "namespace"])
        elif self.is_logical_type(field=field):
            keys_to_ignore.extend(
                [
                    "logicalType",
                ]
            )

        return {name: value for name, value in field.items() if name not in keys_to_ignore}



import dataclasses
import datetime
import decimal
import enum
import inspect
import random
import re
import typing
import uuid

import fastavro
from typing_extensions import get_args, get_origin

from dataclasses_avroschema import (
    exceptions,
    serialization,
    types,
    utils,
    version,
)
from dataclasses_avroschema.faker import fake
from dataclasses_avroschema.utils import is_pydantic_model

from . import field_utils
from .base import Field

if version.PY_VERSION >= (3, 9):  # pragma: no cover
    GenericAlias = (
        typing.GenericAlias,  # type: ignore
        typing._GenericAlias,  # type: ignore
        typing._SpecialGenericAlias,  # type: ignore
        typing._UnionGenericAlias,  # type: ignore
    )  # noqa: E501
else:
    GenericAlias = typing._GenericAlias  # type: ignore  # pragma: no cover


if typing.TYPE_CHECKING:
    from dataclasses_avroschema import AvroModel  # pragma: no cover


__all__ = [
    "ImmutableField",
    "StringField",
    "IntField",
    "LongField",
    "BooleanField",
    "DoubleField",
    "FloatField",
    "BytesField",
    "NoneField",
    "ContainerField",
    "ListField",
    "TupleField",
    "DictField",
    "UnionField",
    "LiteralField",
    "FixedField",
    "EnumField",
    "SelfReferenceField",
    "DateField",
    "DatetimeField",
    "DatetimeMicroField",
    "TimedeltaField",
    "TimeMilliField",
    "TimeMicroField",
    "UUIDField",
    "DecimalField",
    "RecordField",
    "AvroField",
]


class ImmutableField(Field):
    def get_avro_type(
        self,
    ) -> typing.Union[str, typing.List, typing.Dict[str, typing.Any]]:
        if self.default is None:
            return [field_utils.NULL, self.avro_type]
        return self.avro_type


@dataclasses.dataclass
class StringField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.STRING

    def fake(self) -> str:
        return fake.pystr()


@dataclasses.dataclass
class IntField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.INT

    def fake(self) -> int:
        return fake.pyint()


@dataclasses.dataclass
class LongField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.LONG

    def fake(self) -> int:
        return fake.pyint()


@dataclasses.dataclass
class BooleanField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.BOOLEAN

    def fake(self) -> bool:
        return fake.pybool()


@dataclasses.dataclass
class DoubleField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.DOUBLE

    def __post_init__(self):
        super().__post_init__()
        self.extra_default_types_allowed = (int,)

    def fake(self) -> float:
        return fake.pyfloat()


@dataclasses.dataclass
class FloatField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.FLOAT

    def fake(self) -> float:
        return fake.pyfloat()  # Roughly the range on a float32


@dataclasses.dataclass
class BytesField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.BYTES

    def default_to_avro(self, item: bytes) -> str:
        return item.decode()

    def fake(self) -> bytes:
        return fake.pystr().encode()


@dataclasses.dataclass
class NoneField(ImmutableField):
    @property
    def avro_type(self) -> str:
        return field_utils.NULL

    def get_avro_type(self) -> str:
        return self.avro_type


@dataclasses.dataclass
class ContainerField(Field):
    @property
    def avro_type(self) -> typing.Dict:  # type: ignore
        ...  # pragma: no cover

    def get_avro_type(self) -> types.JsonDict:
        avro_type = self.avro_type
        avro_type["name"] = self.inner_name or self.get_singular_name(self.name)

        return avro_type


@dataclasses.dataclass
class BaseListField(ContainerField):
    items_type: typing.Any = None
    internal_field: Field = dataclasses.field(init=False)

    @property
    def avro_type(self) -> typing.Dict:
        self.generate_items_type()
        return {"type": field_utils.ARRAY, "items": self.items_type}

    def get_default_value(self) -> typing.Union[typing.List, dataclasses._MISSING_TYPE]:
        default = super().get_default_value()

        if default is not dataclasses.MISSING:
            if default is None:
                return []
            else:
                if isinstance(default, tuple):
                    default = list(default)
        return default

    def validate_default(self, default):
        assert isinstance(default, list), f"List is required as default for field {self.name}"
        if not isinstance(self.internal_field, UnionField):
            for element in default:
                self.internal_field.validate_default(element)

    def default_to_avro(self, values: typing.List):
        return [self.internal_field.default_to_avro(item) for item in values]

    def generate_items_type(self) -> typing.Any:
        # because avro can have only one type, we take the first one
        items_type = self.type.__args__[0]

        if utils.is_union(items_type):
            self.internal_field = UnionField(
                self.name,
                items_type,
                default=self.default,
                default_factory=self.default_factory,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )
        else:
            self.internal_field = AvroField(
                self.name,
                items_type,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )

        self.items_type = self.internal_field.get_avro_type()


@dataclasses.dataclass
class ListField(BaseListField):
    def fake(self) -> typing.List:
        return [self.internal_field.fake()]


@dataclasses.dataclass
class TupleField(BaseListField):
    """
    This behaves on the same way as `ListField` with
    as in avro schema does not exist `tuples`

    The reason to have this is to generate a proper `fake`
    """

    def fake(self) -> typing.Tuple:
        return (self.internal_field.fake(),)


@dataclasses.dataclass
class DictField(ContainerField):
    values_type: typing.Any = None
    internal_field: typing.Any = None

    def __post_init__(self) -> None:
        super().__post_init__()
        key_type = self.type.__args__[0]

        if not issubclass(key_type, str):
            raise exceptions.InvalidMap(self.name, key_type)

    @property
    def avro_type(self) -> types.JsonDict:
        self.generate_values_type()
        return {"type": field_utils.MAP, "values": self.values_type}

    def get_default_value(
        self,
    ) -> typing.Union[types.JsonDict, dataclasses._MISSING_TYPE]:
        default = super().get_default_value()

        if default is not dataclasses.MISSING:
            if default is None:
                return {}
            else:
                assert isinstance(default, dict), f"Dict is required as default for field {self.name}"
        return default

    def validate_default(self, default):
        assert isinstance(default, dict)
        if not isinstance(self.internal_field, UnionField):
            for element in default.values():
                self.internal_field.validate_default(element)

    def default_to_avro(self, default: typing.Dict):
        return {key: self.internal_field.default_to_avro(value) for key, value in default.items()}

    def generate_values_type(self) -> typing.Any:
        """
        Process typing.Dict. Avro assumes that the key of a map is always a string,
        so we take the second argument to determine the value type
        """
        values_type = self.type.__args__[1]
        self.internal_field = AvroField(
            self.name,
            values_type,
            model_metadata=self.model_metadata,
            parent=self.parent,
        )
        self.values_type = self.internal_field.get_avro_type()

    def fake(self) -> typing.Dict[str, typing.Any]:
        # return a dict of one element with the items type specified
        return {fake.pystr(): self.internal_field.fake()}


@dataclasses.dataclass
class UnionField(Field):
    unions: typing.List = dataclasses.field(default_factory=list)
    internal_fields: typing.List[Field] = dataclasses.field(default_factory=list)
    elements: typing.Tuple = dataclasses.field(default_factory=tuple)

    def generate_unions_type(self) -> typing.List:
        """
        Generate union.

        Arguments:
            elements (typing.List): List of python types
            default (typing.Any): Default value
            default factory (typing.Calleable): Callable to get the default value for
                a list or dict type

        Returns:
            typing.List: List of avro types
        """
        self.elements = get_args(self.type)
        name = self.get_singular_name(self.name)

        unions: typing.List = []

        # Place default at front of list
        default_type = default_field = None
        if self.default is None and self.default_factory is dataclasses.MISSING:
            unions.insert(0, field_utils.NULL)
        elif type(self.default) is not dataclasses._MISSING_TYPE:
            default_type = type(self.default)
            default_field = AvroField(
                name,
                default_type,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )
            unions.append(default_field.get_avro_type())
            self.internal_fields.append(default_field)

        for element in self.elements:
            # create the field and get the avro type
            field = AvroField(name, element, model_metadata=self.model_metadata, parent=self.parent)
            avro_type = field.get_avro_type()

            if avro_type not in unions and field != default_field:
                unions.append(avro_type)
                self.internal_fields.append(field)

        return unions

    def get_avro_type(self) -> typing.List:
        self.unions = self.generate_unions_type()
        return self.unions

    def default_to_avro(self, default: typing.Any):
        first_type = self.internal_fields[0]
        return first_type.default_to_avro(default)

    def validate_default(self, default):
        first_type = self.internal_fields[0]
        first_type.validate_default(default)

    def fake(self) -> typing.Any:
        # get a random internal field and return a fake value
        field = random.choice(self.internal_fields)
        return field.fake()


@dataclasses.dataclass
class LiteralField(Field):
    avro_field: Field = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        """
        Derives Avro schema type[s] and validation requirements
        """
        super().__post_init__()
        args = get_args(self.type)

        # always convert to an Union. If there is Literal has 1 argument then
        # the Union will handle it properly.
        native_type = typing.Union[tuple(type(a) for a in args)]  # type: ignore

        if native_type is str and self.model_metadata.convert_literal_to_enum:
            # if all args are string then we use a `StrEnum` as native type
            native_type = enum.Enum(self.name, {member: member for member in args})  # type: ignore

        self.avro_field = AvroField(
            name=self.name,
            native_type=native_type,
            parent=self.parent,
            default=self.default,
            default_factory=self.default_factory,
            model_metadata=self.model_metadata,
        )

    def get_avro_type(self) -> types.JsonDict:
        return self.avro_field.get_avro_type()

    def default_to_avro(self, default: typing.Any):
        return self.avro_field.default_to_avro(default)

    def validate_default(self, default: typing.Any) -> bool:
        return self.avro_field.validate_default(default)

    def fake(self) -> typing.Any:
        return random.choice(get_args(self.type))


@dataclasses.dataclass
class FixedField(BytesField):
    field_info: typing.Optional[types.FixedFieldInfo] = None
    size: int = 0
    aliases: typing.Optional[typing.List[str]] = None
    namespace: typing.Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self.set_fixed()

    def set_fixed(self) -> None:
        if self.field_info is not None:
            self.size = self.field_info.size
            self.aliases = self.field_info.aliases
            self.namespace = self.field_info.namespace

    def get_avro_type(self) -> types.JsonDict:
        avro_type = {
            "type": field_utils.FIXED,
            "name": self.inner_name or self.get_singular_name(self.name),
            "size": int(self.size),
        }

        if self.namespace is not None:
            avro_type["namespace"] = self.namespace

        if self.aliases is not None:
            avro_type["aliases"] = self.aliases

        return avro_type

    def validate_default(self, default) -> bool:
        msg = "Invalid default type. Default should be bytes"
        assert isinstance(default, bytes), msg
        return True

    def fake(self) -> bytes:
        return fake.pystr(max_chars=self.size).encode()


@dataclasses.dataclass
class EnumField(Field):
    SYMBOL_REGEX: typing.ClassVar = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
    symbols: typing.List[str] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.extra_default_types_allowed = (str,)
        self.symbols = self.get_symbols()
        self.validate_symbols()

    def _get_meta_class_attributes(self) -> typing.Dict[str, typing.Any]:
        # get Enum members
        members = self.type.__members__
        meta = members.get("Meta") or getattr(self.type, "Meta", None)

        doc: typing.Optional[str] = self.type.__doc__

        # On python < 3.11 Enums have a default documentation so we remove it
        if version.PY_VERSION < (3, 11) and doc == "An enumeration.":
            doc = None

        try:
            if meta is not None:
                meta = meta.value
        except AttributeError:
            pass

        metadata = utils.FieldMetadata.create(meta)
        if doc is not None:
            metadata.doc = doc.strip()

        return metadata.to_dict()

    def get_symbols(self) -> typing.List[str]:
        return [member.value for member in self.type if member.name != "Meta"]

    def validate_symbols(self) -> bool:
        for symbol in self.symbols:
            if not isinstance(symbol, str) or not self.SYMBOL_REGEX.fullmatch(symbol):
                raise exceptions.InvalidSymbol(field_name=self.name, symbol=symbol)

        return True

    def get_avro_type(self) -> typing.Union[str, types.JsonDict]:
        metadata = self._get_meta_class_attributes()
        # If the enum name in avro schema different than the python enum class name, use the schema_name.
        name = metadata.pop("schema_name", self.type.__name__)

        if not self.exist_type():
            user_defined_type = utils.UserDefinedType(name=name, model=self.type)
            self.parent._user_defined_types.add(user_defined_type)
            return {
                "type": field_utils.ENUM,
                "name": name,
                "symbols": self.symbols,
                **metadata,
            }
        else:
            namespace = metadata.get("namespace")
            if namespace is None:
                return name
            else:
                return f"{namespace}.{name}"

    def default_to_avro(self, default: typing.Any) -> typing.Any:
        if isinstance(default, str):
            return default
        return default.value

    def fake(self) -> typing.Any:
        return random.choice(self.get_symbols())


@dataclasses.dataclass
class SelfReferenceField(Field):
    def get_avro_type(self) -> typing.Union[typing.List[str], str]:
        str_type = self._get_self_reference_type(self.type)

        if self.default is None:
            # means that default value is None
            return [field_utils.NULL, str_type]
        return str_type

    def get_default_value(self) -> typing.Union[dataclasses._MISSING_TYPE, None]:
        # Only check for None because self reference default value can be only None
        if self.default is None:
            return None
        return dataclasses.MISSING

    def fake(self) -> typing.Any:
        if getattr(self.type, "__args__", None):
            # It means that self.type is `typing.Type['AType']`, and the argument is a string
            # then we return None
            return None
        return self.type.fake()


@dataclasses.dataclass
class DateField(ImmutableField):
    """
    The date logical type represents a date within the calendar,
    with no reference to a particular time zone or time of day.

    A date logical type annotates an Avro int, where the int stores
    the number of days from the unix epoch, 1 January 1970 (ISO calendar).
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATE

    def default_to_avro(self, date: datetime.date) -> int:
        """
        Convert to datetime and get the amount of days
        from the unix epoch, 1 January 1970 (ISO calendar)
        for a given date

        Arguments:
            date (datetime.date)

        Returns:
            int
        """
        date_time = datetime.datetime.combine(date, datetime.datetime.min.time())
        ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts / (3600 * 24))

    def fake(self) -> datetime.date:
        return fake.date_object()


@dataclasses.dataclass
class TimeMilliField(ImmutableField):
    """
    The time-millis logical type represents a time of day,
    with no reference to a particular calendar,
    time zone or date, with a precision of one millisecond.

    A time-millis logical type annotates an Avro int,
    where the int stores the number of milliseconds after midnight, 00:00:00.000.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_TIME_MILIS

    def default_to_avro(self, time: datetime.time) -> int:
        """
        Returns the number of milliseconds after midnight, 00:00:00.000
        for a given time object

        Arguments:
            time (datetime.time)

        Returns:
            int
        """
        hour, minutes, seconds, microseconds = (
            time.hour,
            time.minute,
            time.second,
            time.microsecond,
        )

        return int((((hour * 60 + minutes) * 60 + seconds) * 1000) + (microseconds / 1000))

    def fake(self) -> datetime.time:
        return fake.time_object()


@dataclasses.dataclass
class TimeMicroField(ImmutableField):
    """
    The time-micros logical type represents a time of day,
    with no reference to a particular calendar,
    time zone or date, with a precision of one millisecond.

    A time-micros logical type annotates an Avro long,
    where the int stores the number of milliseconds after midnight, 00:00:00.000000.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_TIME_MICROS

    def default_to_avro(self, time: datetime.time) -> float:
        """
        Returns the number of microseconds after midnight, 00:00:00.000000
        for a given time object

        Arguments:
            time (datetime.time)

        Returns:
            int
        """
        hour, minutes, seconds, microseconds = (
            time.hour,
            time.minute,
            time.second,
            time.microsecond,
        )

        return int((((hour * 60 + minutes) * 60 + seconds) * 1000000) + microseconds)

    def fake(self) -> datetime.time:
        datetime_object: datetime.datetime = fake.date_time(tzinfo=datetime.timezone.utc)
        datetime_object = datetime_object + datetime.timedelta(microseconds=random.randint(0, 999))
        return datetime_object.time()


@dataclasses.dataclass
class TimedeltaField(ImmutableField):
    """
    The timedelta logical represents a absolute length of time.

    It annotates an Avro `double`, which stores a number of seconds to microsecond precision.
    Note that is different than an Avro `duration`, which could represent different lengths of time depending on
    when it is measured from.

    This is not an official Avro logical type, so consumers will need to know how to handle it.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_TIMEDELTA

    def default_to_avro(self, value: datetime.timedelta) -> float:
        return self.to_avro(value)

    @classmethod
    def from_avro(cls, value: float) -> datetime.timedelta:
        """Convert from a fastavro-supported type to a timedelta."""
        return datetime.timedelta(seconds=value)

    @classmethod
    def to_avro(cls, value: datetime.timedelta) -> float:
        """Convert from a timedelta to a fastavro-supported type."""
        return value.total_seconds()

    def fake(self) -> datetime.timedelta:
        return fake.time_delta(end_datetime=fake.date_time())


def _fastavro_serialize_timedelta(data: typing.Any, *_) -> typing.Any:
    """Serialize a timedelta for fastavro."""
    # Fastavro will call this function on any value that is provided for a
    # union type where one of the union members is a timedelta, so we have to
    # explicitly check if the value is a timedelta before trying to convert it.
    if isinstance(data, datetime.timedelta):
        return TimedeltaField.to_avro(data)
    return data


def _fastavro_deserialize_timedelta(data: float, *_) -> datetime.timedelta:
    """Deserialize a timedelta for fastavro."""
    return TimedeltaField.from_avro(data)


fastavro.write.LOGICAL_WRITERS["double-dataclasses-avroschema-timedelta"] = _fastavro_serialize_timedelta
fastavro.read.LOGICAL_READERS["double-dataclasses-avroschema-timedelta"] = _fastavro_deserialize_timedelta


@dataclasses.dataclass
class DatetimeField(ImmutableField):
    """
    The timestamp-millis logical type represents an instant on the global timeline,
    independent of a particular time zone or calendar, with a precision of one millisecond.

    A timestamp-millis logical type annotates an Avro long,
    where the long stores the number of milliseconds from the unix epoch,
    1 January 1970 00:00:00.000 UTC.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATETIME_MILIS

    def default_to_avro(self, date_time: datetime.datetime) -> int:
        """
        Returns the number of milliseconds from the unix epoch,
        1 January 1970 00:00:00.000 UTC for a given datetime
        """
        if date_time.tzinfo:
            ts = (date_time - utils.epoch).total_seconds()
        else:
            ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts * 1000)

    def fake(self) -> datetime.datetime:
        return fake.date_time(tzinfo=datetime.timezone.utc)


@dataclasses.dataclass
class DatetimeMicroField(ImmutableField):
    """
    The timestamp-millis logical type represents an instant on the global timeline,
    independent of a particular time zone or calendar, with a precision of one millisecond.

    A timestamp-millis logical type annotates an Avro long,
    where the long stores the number of milliseconds from the unix epoch,
    1 January 1970 00:00:00.000000 UTC.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATETIME_MICROS

    def default_to_avro(self, date_time: datetime.datetime) -> int:
        """
        Returns the number of milliseconds from the unix epoch,
        1 January 1970 00:00:00.000000 UTC for a given datetime
        """
        if date_time.tzinfo:
            ts = (date_time - utils.epoch).total_seconds()
        else:
            ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts * 1000000)

    def fake(self) -> datetime.datetime:
        datetime_object: datetime.datetime = fake.date_time(tzinfo=datetime.timezone.utc)
        return datetime_object + datetime.timedelta(microseconds=random.randint(0, 999))


@dataclasses.dataclass
class UUIDField(ImmutableField):
    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_UUID

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {str} or {uuid.UUID}"
        assert isinstance(default, (str, uuid.UUID)), msg

        return True

    def default_to_avro(self, uuid: uuid.UUID) -> str:
        return str(uuid)

    def fake(self) -> uuid.UUID:
        return uuid.uuid4()


@dataclasses.dataclass
class DecimalField(Field):
    field_info: typing.Optional[types.DecimalFieldInfo] = None
    max_digits: int = -1  # amount of digits, in avro is called `precision`
    decimal_places: int = 0  # amount of digits to the right side, in avro is called `scale`

    def __post_init__(self) -> None:
        super().__post_init__()
        self.set_precision_scale()

    def set_precision_scale(self) -> None:
        if self.field_info is not None:
            self.max_digits = self.field_info.max_digits
            self.decimal_places = self.field_info.decimal_places

        # Validation on precision and scale per Avro schema
        if self.max_digits <= 0:
            raise ValueError("`max_digits` must be a positive integer greater than zero")

        if self.decimal_places < 0 or self.max_digits < self.decimal_places:
            raise ValueError("`decimal_places` must be zero or a positive integer less than or equal to the precision.")

    def get_avro_type(
        self,
    ) -> typing.Union[types.JsonDict, typing.List[typing.Union[str, types.JsonDict]]]:
        avro_type = {
            "type": field_utils.BYTES,
            "logicalType": field_utils.DECIMAL,
            "precision": self.max_digits,
            "scale": self.decimal_places,
        }

        if self.default is None:
            return ["null", avro_type]

        return avro_type

    def default_to_avro(self, default: decimal.Decimal) -> str:
        return serialization.decimal_to_str(default, self.max_digits, self.decimal_places)

    def fake(self) -> decimal.Decimal:
        return fake.pydecimal(
            right_digits=self.decimal_places,
            left_digits=self.max_digits - self.decimal_places,
        )


@dataclasses.dataclass
class RecordField(Field):
    def get_avro_type(self) -> typing.Union[str, typing.List, typing.Dict]:
        meta = getattr(self.type, "Meta", type)
        metadata = utils.SchemaMetadata.create(meta)

        if self.parent is not None and self.parent._parser is not None:
            alias = self.parent._parser.metadata.get_alias_nested_items(self.name)
        else:
            alias = metadata.get_alias_nested_items(self.name)  # type: ignore  # noqa E501

        # The priority for the schema name
        # 1. Check if exists an alias_nested_items in parent llass or Meta class of own model
        # 2. Check if the schema_name is present in the Meta class of own model
        # 3. Use the default class Name (self.type.__name__)
        name = alias or metadata.schema_name or self.type.__name__

        if not self.exist_type() or alias is not None:
            user_defined_type = utils.UserDefinedType(name=name, model=self.type)
            self.parent._user_defined_types.add(user_defined_type)

            record_type = self.type.avro_schema_to_python(parent=self.parent)
            record_type["name"] = name
        else:
            if metadata.namespace is None:
                record_type = name
            else:
                record_type = f"{metadata.namespace}.{name}"

        if self.default is None:
            return [field_utils.NULL, record_type]

        return record_type

    def default_to_avro(self, value: "AvroModel") -> typing.Dict:
        parser = value._parser or value._generate_parser()
        return {
            fieldname: field.default_to_avro(getattr(value, fieldname))
            for fieldname, field in parser.get_fields_map().items()
        }

    def fake(self) -> typing.Any:
        return self.type.fake()


from .mapper import (
    ALL_TYPES_FIELD_CLASSES,
    CONTAINER_FIELDS_CLASSES,
    IMMUTABLE_FIELDS_CLASSES,
    LOGICAL_TYPES_FIELDS_CLASSES,
    SPECIAL_ANNOTATED_TYPES,
)

LOGICAL_CLASSES = LOGICAL_TYPES_FIELDS_CLASSES.keys()
PYDANTIC_CUSTOM_CLASS_METHOD_NAMES = {
    "__get_validators__",
    "__get_pydantic_core_schema__",
}


def field_factory(
    name: str,
    native_type: typing.Any,
    parent: typing.Optional[typing.Type["AvroModel"]] = None,
    *,
    default: typing.Any = dataclasses.MISSING,
    default_factory: typing.Any = dataclasses.MISSING,
    metadata: typing.Optional[typing.Dict[str, typing.Any]] = None,
    model_metadata: typing.Optional[utils.SchemaMetadata] = None,
) -> Field:
    from dataclasses_avroschema import AvroModel

    if parent is None:
        # if parent is None, then we assume that the field is defined in an AvroModel
        # and we set the parent to AvroModel
        parent = AvroModel

    if model_metadata is None:
        model_metadata = utils.SchemaMetadata()

    if metadata is None:
        metadata = {}

    field_info = None
    if native_type is None:
        native_type = type(None)

    if utils.is_annotated(native_type):
        a_type, *extra_args = get_args(native_type)
        field_info = next((arg for arg in extra_args if isinstance(arg, types.FieldInfo)), None)

        if field_info is not None:
            # it means that it is a custom type defined by us `Int32`, `Float32`,`TimeMicro`, `DateTimeMicro`
            # confixed or condecimal
            native_type = utils.rebuild_annotation(a_type, field_info)

        if native_type not in ALL_TYPES_FIELD_CLASSES:
            # type Annotated with the end user
            native_type = a_type

    if native_type in IMMUTABLE_FIELDS_CLASSES:
        klass = IMMUTABLE_FIELDS_CLASSES[native_type]
        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )

    # special case for some dynamic pydantic types (especially constraint types)
    # when a type cannot be imported and needs to be referenced by qualified string
    # see pydantic conint() implementation for more information
    elif inspect.isclass(native_type) and f"{native_type.__name__}" in IMMUTABLE_FIELDS_CLASSES:
        klass = IMMUTABLE_FIELDS_CLASSES[f"{native_type.__name__}"]
        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )

    elif utils.is_self_referenced(native_type, parent):
        return SelfReferenceField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif native_type in (types.Fixed, decimal.Decimal):
        klass = SPECIAL_ANNOTATED_TYPES[native_type]  # type: ignore
        return klass(  # type: ignore
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
            field_info=field_info,
        )
    elif native_type in LOGICAL_TYPES_FIELDS_CLASSES:
        klass = LOGICAL_TYPES_FIELDS_CLASSES[native_type]  # type: ignore

        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif isinstance(native_type, GenericAlias):  # type: ignore
        origin = get_origin(native_type)

        if origin in CONTAINER_FIELDS_CLASSES:
            container_klass = CONTAINER_FIELDS_CLASSES[origin]
        elif origin is typing.Literal:
            container_klass = LiteralField
        else:
            raise ValueError(
                f"Invalid Type {native_type} for field {name}. "
                "Accepted types are list, tuple, dict, typing.Union, or typing.Literal"
            )

        # check here
        return container_klass(  # type: ignore
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif inspect.isclass(native_type) and issubclass(native_type, enum.Enum):
        return EnumField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif types.UnionType is not None and isinstance(native_type, types.UnionType):
        # we need to check whether types.UnionType because it works only in
        # python 3.9 or importing __future__ in previous python versions
        # cases when a container is used, for example `typing.List[int] | str` in python is
        # translated to typing.Union[typing.List[int], str] so it won't reach this point
        return UnionField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif inspect.isclass(native_type) and issubclass(native_type, AvroModel):
        return RecordField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    # See if this is a pydantic "Custom Class"
    elif (
        inspect.isclass(native_type)
        and not is_pydantic_model(native_type)  # type: ignore[arg-type]
        and any(method_name in dir(native_type) for method_name in PYDANTIC_CUSTOM_CLASS_METHOD_NAMES)
    ):
        if getattr(parent, "__config__", None):
            try:
                # Build a field for the encoded type since that's what will be serialized
                encoded_type = parent.__config__.json_encoders[native_type]
            except KeyError:
                raise ValueError(
                    f"Type {native_type} for field {name} must be "
                    "listed in the pydantic 'json_encoders' config for {parent}"
                    " (or for one of the classes in its inheritance tree since "
                    "pydantic configs are inherited)"
                )
        else:
            encoded_type = parent.model_config["json_encoders"][native_type]

        # default_factory is not schema-friendly for Custom Classes since it could be returning
        # dynamically constructed values that should not be treated as defaults. For example,
        # native_type could be a timestamp field with a default factory that returns the current
        # time, which would result in "generate_schema" producing a schema with a different default
        # everytime that it's called from AvroModel.
        default_factory = dataclasses.MISSING

        # Encode the default value if it's an instance of native_type
        default_value = encoded_type(default) if type(default) is native_type else default

        # Build a field for the encoded type
        return field_factory(
            name,
            native_type=encoded_type,
            parent=parent,
            default=default_value,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
        )
    else:
        msg = (
            f"Type {native_type} for field {name} is unknown. Please check the valid types at "
            "https://marcosschroh.github.io/dataclasses-avroschema/fields_specification/#avro-field-and-python-types-summary"  # noqa: E501
        )

        raise ValueError(msg)


AvroField = field_factory



import json
from typing import Any, Dict, Type

from fastavro.validation import validate

from dataclasses_avroschema import AvroModel, serialization
from dataclasses_avroschema.types import JsonDict
from dataclasses_avroschema.utils import standardize_custom_type

from .parser import PydanticParser

try:
    from pydantic import BaseModel  # pragma: no cover
except ImportError as ex:  # pragma: no cover
    raise Exception("pydantic must be installed in order to use AvroBaseModel") from ex  # pragma: no cover


class AvroBaseModel(BaseModel, AvroModel):  # type: ignore
    @classmethod
    def json_schema(cls: Type["AvroBaseModel"], *args: Any, **kwargs: Any) -> str:
        return json.dumps(cls.model_json_schema(*args, **kwargs))

    def _standardize_type(self) -> Dict[str, Any]:
        """
        Standardization factory that converts data according to the
        user-defined pydantic json_encoders prior to passing values
        to the standard type conversion factory
        """
        encoders = self.model_config.get("json_encoders") or {}
        data = dict(self)

        for k, v in data.items():
            v_type = type(v)
            if v_type in encoders:
                encode_method = encoders[v_type]
                data[k] = encode_method(v)
        return data

    def asdict(self) -> JsonDict:
        """
        Returns this model in dictionary form. This method differs from
        pydantic's dict by converting all values to their Avro representation.
        It also doesn't provide the exclude, include, by_alias, etc.
        parameters that dict provides.
        """
        return {
            field_name: standardize_custom_type(
                field_name=field_name, value=field_value, model=self, base_class=AvroBaseModel
            )
            for field_name, field_value in self._standardize_type().items()
        }

    @classmethod
    def parse_obj(cls: Type["AvroBaseModel"], data: Dict) -> "AvroBaseModel":
        return cls.model_validate(obj=data)

    def to_dict(self) -> JsonDict:
        return self.model_dump()

    def serialize(self, serialization_type: serialization.SerializationType = "avro") -> bytes:
        """
        Overrides the base AvroModel's serialize method to inject this
        class's standardization factory method
        """
        schema = self.avro_schema_to_python()

        return serialization.serialize(
            self.asdict(),
            schema,
            serialization_type=serialization_type,
        )

    def validate_avro(self) -> bool:
        """
        Validate that instance matches the avro schema
        """
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    @classmethod
    def fake(cls: Type["AvroBaseModel"], **data: Any) -> "AvroBaseModel":
        """
        Creates a fake instance of the model.

        Attributes:
            data: Dict[str, Any] represent the user values to use in the instance
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return cls.model_validate(payload)

    @classmethod
    def _generate_parser(cls: Type["AvroBaseModel"]) -> PydanticParser:
        return PydanticParser(type=cls, parent=cls._parent or cls)



from typing import Any, Dict, Type

from fastavro.validation import validate

from dataclasses_avroschema import AvroModel, serialization
from dataclasses_avroschema.types import JsonDict
from dataclasses_avroschema.utils import standardize_custom_type

from .parser import PydanticV1Parser

try:
    from pydantic.v1 import BaseModel  # pragma: no cover
except ImportError as ex:  # pragma: no cover
    raise Exception("pydantic must be installed in order to use AvroBaseModel") from ex  # pragma: no cover


class AvroBaseModel(BaseModel, AvroModel):  # type: ignore
    @classmethod
    def json_schema(cls: Type["AvroBaseModel"], *args: Any, **kwargs: Any) -> str:
        return cls.schema_json(*args, **kwargs)

    def _standardize_type(self) -> Dict[str, Any]:
        """
        Standardization factory that converts data according to the
        user-defined pydantic json_encoders prior to passing values
        to the standard type conversion factory
        """
        encoders = self.__config__.json_encoders
        data = dict(self)

        for k, v in data.items():
            v_type = type(v)
            if v_type in encoders:
                encode_method = encoders[v_type]
                data[k] = encode_method(v)
        return data

    def asdict(self) -> JsonDict:
        """
        Returns this model in dictionary form. This method differs from
        pydantic's dict by converting all values to their Avro representation.
        It also doesn't provide the exclude, include, by_alias, etc.
        parameters that dict provides.
        """
        return {
            field_name: standardize_custom_type(
                field_name=field_name, value=value, model=self, base_class=AvroBaseModel
            )
            for field_name, value in self._standardize_type().items()
        }

    def to_dict(self) -> JsonDict:
        return dict(self)

    def serialize(self, serialization_type: serialization.SerializationType = "avro") -> bytes:
        """
        Overrides the base AvroModel's serialize method to inject this
        class's standardization factory method
        """
        schema = self.avro_schema_to_python()

        return serialization.serialize(
            self.asdict(),
            schema,
            serialization_type=serialization_type,
        )

    def validate_avro(self) -> bool:
        """
        Validate that instance matches the avro schema
        """
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    @classmethod
    def fake(cls: Type["AvroBaseModel"], **data: Any) -> "AvroBaseModel":
        """
        Creates a fake instance of the model.

        Attributes:
            data: Dict[str, Any] represent the user values to use in the instance

        Returns:
            AvroModel instance
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return cls.parse_obj(payload)

    @classmethod
    def _generate_parser(cls: Type["AvroBaseModel"]) -> PydanticV1Parser:
        return PydanticV1Parser(type=cls, parent=cls._parent or cls)
